'''
Pytorch implementation of wide resnet.

Reference:
[1] S. Zagoruyko and N. Komodakis. Wide residual networks. arXiv preprint arXiv:1605.07146, 2016.
'''

import torch
import torch.nn as nn
import math


def conv3x3(in_planes, out_planes, stride=1):
    " 3x3 convolution with padding "
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion=1

    def __init__(self, inplanes, planes, stride=1, downsample=None, outdim=0):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        if outdim == 0:
            self.conv2 = conv3x3(planes, planes)
            self.bn2 = nn.BatchNorm2d(planes)
        else:
            self.conv2 = conv3x3(planes, outdim)
            self.bn2 = nn.BatchNorm2d(outdim)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Wide_ResNet(nn.Module):

    def __init__(self, block, layers, wfactor, num_classes=10, fc_bias = True, fixdim= False, ETF_fc = False,):
        super(Wide_ResNet, self).__init__()
        self.inplanes = 16
        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(block, 16*wfactor, layers[0])
        self.layer2 = self._make_layer(block, 32*wfactor, layers[1], stride=2)
        if not fixdim:
            self.layer3 = self._make_layer(block, 64*wfactor, layers[2], stride=2)
            self.avgpool = nn.AvgPool2d(8, stride=1)
            self.fc = nn.Linear(64 * block.expansion * wfactor, num_classes)
        else:
            self.layer3 = self._make_layer(block, 64*wfactor, layers[2], stride=2, outdim=fixdim)
            self.avgpool = nn.AvgPool2d(8, stride=1)
            self.fc = nn.Linear(fixdim, num_classes, bias=fc_bias)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                if ETF_fc:
                    weight = torch.sqrt(torch.tensor(num_classes/(num_classes-1)))*(torch.eye(num_classes)-(1/num_classes)*torch.ones((num_classes, num_classes)))
                    weight /= torch.sqrt((1/num_classes*torch.norm(weight, 'fro')**2))
                    if fixdim:
                        m.weight = nn.Parameter(torch.mm(weight, torch.eye(num_classes, fixdim)))
                    else:
                        m.weight = nn.Parameter(torch.mm(weight, torch.eye(num_classes, 64*wfactor)))
                    m.weight.requires_grad_(False)

    def _make_layer(self, block, planes, blocks, stride=1, outdim=0):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion)
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for _ in range(1, int(blocks)-1):
            layers.append(block(self.inplanes, planes))

        downsample = None
        if outdim != 0:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, outdim, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(outdim)
            )

        layers.append(block(self.inplanes, planes, downsample=downsample, outdim=outdim))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        x = self.avgpool(x)
        feature = x.view(x.size(0), -1)
        x = self.fc(feature) / self.temp

        return x, feature


def wide_resnet(num_classes=10, depth=26, width=10, **kwargs):
    assert (depth - 2) % 6 == 0
    n = (depth - 2) / 6
    return Wide_ResNet(BasicBlock, [n, n, n], width, num_classes=num_classes, **kwargs)